
/*******************************************************************************
*
*                                 thermNeon
*
********************************************************************************
* File name:        thermneon_main.c
* Dependencies:     See includes section below
* Processor:        PIC18F1330
* Hardware:         thermneon v3
* Compiler:         C18 v3.37.01
* IDE:              MPLABX v1.10
********************************************************************************
* File description: thermNeon firmware source code
*
* Revision:         v2.1 - 29/07/14
*
* Author:           SALTechips - Alex Tsekenis
* Notes:            -For thermNeons using the remote sensor mod
 *                  -Not to be used with HW V1 or V2
*                   -1-wire and EEPROM libraries are included here.
********************************************************************************
* Licence:
*
* Copyright (C) 2014 SALTechips
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.

* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see http://www.gnu.org/licenses
*******************************************************************************/

#include <p18f1330.h>   //Include device header file
#include <delays.h>     //Include delays library header file
#include <math.h>       //Include math library

// <editor-fold defaultstate="collapsed" desc="Device configuration bits">
#pragma config OSC = INTIO2 //Internal RC, both RA6 and RA7 as GPIO
#pragma config FCMEN = OFF  //Fail-Safe Clock Monitor disabled
#pragma config IESO = OFF   //Internal External Switchover mode disabled
#pragma config PWRT = ON    //Power-up timer enabled
#pragma config BOR = OFF    //Brown-out Reset disabled
#pragma config WDT = OFF    //WDT disabled
#pragma config PWMPIN = OFF //PWM outputs disabled upon Reset
#pragma config MCLRE = ON   //MCLR pin enabled, RA5 input pin disabled
#pragma config STVREN = OFF //Stack full/underflow will not cause Reset
#pragma config XINST = OFF  //Instruction set extension and Indexed Addressing mode disabled
#pragma config DEBUG = OFF  //Background debugger disabled, RB6 and RB7 configured as general purpose I/O pins
#pragma config CP0 = OFF    //Block 0 is not code-protected
#pragma config CP1 = OFF    //Block 1 is not code-protected
#pragma config CPB = OFF    //Boot block is not code-protected
#pragma config CPD = OFF    //EEPROM is not code protected
#pragma config WRTD = OFF   //Data EEPROM is not write-protected
#pragma config WRT0 = OFF   //Block 0 is not write-protected
#pragma config WRT1 = OFF   //Block 1 is not write-protected
#pragma config WRTC = OFF   //Configuration registers are not write-protected
#pragma config WRTB = OFF   //Boot Block is not write-protected
#pragma config EBTR0 = OFF  //Block 0 is not protected from table reads executed in other blocks
#pragma config EBTR1 = OFF  //Block 1 is not protected from table reads executed in other blocks
#pragma config EBTRB = OFF  //Boot Block is not protected from table reads executed in other blocks
// </editor-fold>

// <editor-fold defaultstate="collapsed" desc="EEPROM data loading">
#pragma romdata eeprom_data=0xf00000
rom unsigned char eedata_values[]={0x01,0x3C,0x23,0x0F,0x68,0x01,0xF0,0x00,0x64,0x00,0x10,0x27}; //See thermneon EEPROM default configuration
#pragma code
// </editor-fold>

// <editor-fold defaultstate="collapsed" desc="Definitions of symbolic names">
#define PB_SET  PORTAbits.RA0       //Pushbutton Set
#define PB_UP   PORTAbits.RA1       //Pushbutton Up
#define OW_PORT PORTAbits.RA6       //Temperature sensor. This is used as a bidirectional read/write register for 1-wire communications. A write to the PORT is transformed into a write to the LAT register.
#define OW_RW  TRISAbits.RA6        //Direction control register for 1-wire communications. Also see direction definitions below.
enum OW_DIRECTION {WRITE, READ};    //Direction definitions. WRITE is 1 and READ is 0. When assigned to T_SENSE_RW the correct direction (input/output) is set for the temperature sensor pin.
#define OW_SEARCH_ROM       0xF0    //1-wire bus DS18B20 commands
#define OW_READ_ROM         0x33
#define OW_MATCH_ROM        0x55
#define OW_SKIP_ROM         0xCC
#define OW_ALARM_SEARCH     0xEC
#define OW_CONVERT_T        0x44
#define OW_WRITE_PAD        0x4E
#define OW_READ_PAD         0xBE
#define OW_COPY_PAD         0x48
#define OW_RECALL_EEPROM    0xB8
#define OW_READ_SUPPY       0xB4
#define LED_G   LATBbits.LATB4      //Green LED
#define LED_R   LATBbits.LATB0
#define LED_B   LATAbits.LATA7      
enum MENU_FLASH_COLOURS {RED, GREEN};   //Menu flash led colours
enum PWM_GEN {BRIGHT, HV, BAR};     //PWM generator definitions. BRIGHT is for global brightness control of the LEDs, HV for the HV boost converter, BAR for the bargraph length
//#define LED_SW        LATBbits.LATB1   //Global LED ON/OFF switch. Normally unused as the pin is under HW PWM control.
//#define HV_BOOST      LATBbits.LATB5   //HV boost converter control. Normally unused as the pin is under HW PWM control.
//#define BARGRAPH_DRIVE LATBbits.LATB7  //Bargraph control. Normally unused as the pin is under HW PWM control.
//#define HV_BOOST_FB   PORTBbits.RB2    //HV boost converter feedback. Normally unused as the pin is connected to an analogue HW comparator. Can be used in digital mode as a Schmitt trigger input.
//#define CHUCK_NORRIS 0xFF              //Because you cannot define Chuck Norris.
enum T1_CONTEXTS {SLEEP, FLASH, PWM};    //All the contexts T1 can operate in.
enum EEPROM_LOC {DISP_FREQ_LOC=0x00, DISP_DUR_LOC=0x01, HOT_SET_LOC=0x02,
COLD_SET_LOC=0x03, HOT_HUE_LOC=0x04, COLD_HUE_LOC=0x06, ILLUM_LVL_LOC=0x08, SPUTTER_LOC=0x0A};    //EEPROM address locations for all user settings.

enum ILLUM_COLOURS {RED_HUE_H=360, RED_HUE_L=0, MAGENTA_HUE=300, BLUE_HUE=240, CYAN_HUE=180, GREEN_HUE=120, YELLOW_HUE=60, WHITE}; //Hue of scale illumination colours. White is a special case.
enum ILLUM_LEVEL    {OFF=0, LVL_1=10, LVL_2=40, LVL_3=100, LVL_4=200, LVL_5=400, MAX=1024}; //Levels of scale illumination and associated PWM register values
// </editor-fold>

// <editor-fold defaultstate="collapsed" desc="Function prototypes">
void low_isr (void);
void high_isr (void);
void init (void);
void write_PWM_duty(unsigned char gen, unsigned int duty);
unsigned int read_PWM_duty(unsigned char gen);
void HV_soft_start(void);
void wakeup_HV(void);
void sleep_HV(void);
void calibration(void);
void start_flash(unsigned char colour);
void stop_flash(void);
unsigned int temp_to_pwm (unsigned char temp);
void display_temp(unsigned char temp);
void enter_menu(void);
void enter_menu_option(unsigned char option);
unsigned char OW_reset(void);
void OW_write_bit(unsigned char data);
void OW_write_byte(unsigned char data);
unsigned char OW_read_bit(void);
unsigned char OW_read_byte(void);
void convert_temp(void);
unsigned int read_temp(void);
void delay_minx(unsigned char minutes);
void delay_secx(unsigned char seconds);
void sputter (void);
void SaveByteToEEPROM(unsigned char address, unsigned char data);
void SaveIntToEEPROM(unsigned char start_address, unsigned int data);
unsigned char LoadByteFromEEPROM(unsigned char address);
unsigned int LoadIntFromEEPROM(unsigned char start_address);
unsigned char EEPROMRead(unsigned char address);
void EEPROMWrite(unsigned char address, unsigned char data);
void load_user_settings(void);
void save_user_settings(void);
void start_LED_pwm(void);
void stop_LED_pwm(void);
void hue_to_RGB(unsigned int H, unsigned char *R, unsigned char *G, unsigned char *B);
unsigned int abs(signed int value);
unsigned int find_hue (unsigned int hot_hue, unsigned int cold_hue, unsigned char hot_set, unsigned char cold_set, unsigned char temperature);
void prep_scale_colour(void);
// </editor-fold>

// <editor-fold defaultstate="collapsed" desc="Variables">
union Integer{              // create an integer data type
    unsigned int lt;
    char bt[2];             // that contains two bytes. This is used to write/write nteger values into/from byte-long registers
};
unsigned volatile char led_flash_counter=0;     //counter used by Timer 1 to perform flashing of LEDs during menu display. See ISR.
unsigned volatile char led_PWM_counter=0;       //counter used by Timer 1 to perform PWM control of RGB LEDs during temperature display. See ISR.
unsigned volatile int  sleep_counter=0;         //counter used by Timer 1 to perform sleep timing of unit. See ISR.
unsigned volatile char minutes_slept=0;         //counter used by Timer 1 to perform sleep timing of unit. See ISR.
unsigned volatile char pb_set_flag=0;           //Flags to indicate if and which pushbutton has been pressed
unsigned volatile char pb_up_flag=0;
unsigned volatile char T1_context;              //The context Timer 1 operates in.
unsigned char disp_freq;                        //Display frequency setting in minutes
unsigned char disp_dur;                         //Display duration setting in seconds
unsigned char red_duty;                         //duty cycles values for LEDs. 3 bits resolution, 7 is maximum value.
unsigned char green_duty;
unsigned char blue_duty;
unsigned int hot_hue;                           //Hues of colours assigned to hot and cold temperature setpoints
unsigned int cold_hue;
unsigned char hot_set;                          //Temperature setpoints for colour fading
unsigned char cold_set;
unsigned int illum_level;                       //scale illumination level
unsigned char temp;                             //last read temperature . Value is double the actual temperature.
// </editor-fold>

// <editor-fold defaultstate="collapsed" desc="Interrupt Service Routines">

#pragma code low_vetor=0x18     //Place this code at address 0x0018 (low priority interrupt vector)
void low_vector(void)
{
_asm GOTO low_isr _endasm       //Branch to the function low_isr
}
#pragma code	// Return to default code section

#pragma code high_vetor=0x8     //Place this code at address 0x0008 (high priority interrupt vector)
void high_vector(void)
{
_asm GOTO high_isr _endasm      //Branch to the function high_isr
}
#pragma code	// Return to default code section

#pragma interruptlow low_isr    //Low priority interrupt service routine
void low_isr (void)
{
    if(CMCONbits.C2OUT==0){     //Comparator output is low therefore output is below threshold
        PDC1H = 0x03;           //Increase the duty cycle of the HV converter to boost output voltage
        PDC1L = 0x0F;           //Value written is 783 decimal
    }
    else{                       //Comparator output is high therefore output is above threshold
        PDC1H = 0x01;           //Decrease the duty cycle of the HV converter to drop the output voltage
        PDC1L = 0x40;           //Value written is 320 decimal
   }
    TMR0H=0xFE;                     //Reset the timer. Value written is 65115.
    TMR0L=0x5B;
    INTCONbits.TMR0IF=0;            //Clear Timer 0 interrupt flag
    //}
}

#pragma interrupt high_isr       //High priority interrupt service routine
void high_isr (void)
{

    if(PIR1bits.TMR1IF){

        T1CONbits.TMR1ON=0;     //Stop the timer to give time for the ISR to execute

        switch(T1_context){     //Depending on the current context, execute the correct code segment
            case PWM:           //Timer 1 is used to control PWM of RGB LEDs
                ++led_PWM_counter;
                if (led_PWM_counter<15){                    //During the PWM period
                    if (red_duty==led_PWM_counter)LED_R=0;  //Is it time to switch off a colour?
                    if (green_duty==led_PWM_counter)LED_G=0;
                    if (blue_duty==led_PWM_counter)LED_B=0;
                }
                else{
                    led_PWM_counter=0;                      //After a PWM period, reset the LEDs
                    if(red_duty>0)LED_R=1;
                    if(green_duty>0)LED_G=1;
                    if(blue_duty>0)LED_B=1;
                }
                TMR1H=0xFD;                //Reset the timer, value written is 64910
                TMR1L=0x8E;
                break;
            case FLASH:         //Timer 1 is used to control the flash of the LEDs during menu display
                        ++led_flash_counter;             //Increment counter
                        if(led_flash_counter==8){       //If the counter has reached 8, 524288 cycles or 0.5seconds have lapsed
                            led_flash_counter=0;        //Reset the counter for the next time
                            if((PDC0H==0x02)&&(PDC0L==0xBC)){   //If the PWM register has a value of 700
                                PDC0H=0;                        //Set it to 40 (low brightness)
                                PDC0L=0x28;
                            }
                            else{                       //Otherise set it to 700 (high brightness)
                                PDC0H=0x02;
                                PDC0L=0xBC;
                            }
                        }
                        TMR1H=0;                //Reset the timer
                        TMR1L=0;
                        break;
            case SLEEP:     //Timer 1 is used as an alarm timer during sleep
                        ++sleep_counter;
                        if(sleep_counter==915){ //915 ticks of the sleep counter is one minute of sleep
                            sleep_counter=0;
                            ++minutes_slept;
                        }
                        TMR1H=0;                //Reset the timer
                        TMR1L=0;
                        break;
            default:    break;
        }
        T1CONbits.TMR1ON=1;     //Restart the timer
        PIR1bits.TMR1IF=0;      //Clear Timer 1 interrupt flag
    }

    else{
        if(INTCONbits.INT0IF){      //If the interrupt was from the SET button
            pb_set_flag=0xFF;       //Set the flag for SET pushbutton
            INTCONbits.INT0IF=0;
        }
        else if(INTCON3bits.INT1IF){   //If the interrupt was from the UP button
            pb_up_flag=0xFF;           //Set the flag for UP pushbutton
            INTCON3bits.INT1IF=0;
        }
    }

}
// </editor-fold>

/************************************************
* Function:		main
* Revision - Date:	3 - 29/07/2014
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	None
* Output:		None
* Side effects:		None
* Overview:		Main program entry point.
*
* Note:			None
*************************************************/
void main (void){

    unsigned int i;

    init();                         //Initialise unit

    load_user_settings();           //load user configuration settings

    while(1){

            T1CONbits.TMR1ON=0;     //Wake up, stop the sleep timer
            wakeup_HV();            //Prepare the HV
            pb_up_flag=0;           //In case this was the reason to wake up
            if(pb_set_flag){        //If SET button was the reason to wake up,
                pb_set_flag=0;      //Clear the flag
                enter_menu();       //Enter configuration menu
                Delay10KTCYx(255);  //Wait a bit before displaying the temperature with the new settings
                Delay10KTCYx(255);
            }

            convert_temp();             //Initialise temperature conversion
            temp=read_temp();           //Read converted temperature
            
            prep_scale_colour();            //Set the right RGB colour to the LEDs
            start_LED_pwm();                //Start PWM control of LEDs

            for(i=0;i<illum_level;i++){     //Ramp up scale illumination
              write_PWM_duty(BRIGHT,i);
              Delay1KTCYx(20);
            }
            for(i=0;i<temp_to_pwm(temp);i++){   //Ramp up bargraph
              write_PWM_duty(BAR,i);
              Delay1KTCYx(50);
            }

            if(PB_UP)                       //At this point, if the UP pushbutton is not pressed
                delay_secx(disp_dur);       //Wait for set display time
            else                            //Otherwise
                for (i=0;i<270;i++){        //Repeat 270 times, approximately 30 seconds
                    convert_temp();         //Initialise temperature conversion
                    temp=read_temp();       //Read converted temperature
                    prep_scale_colour();    //Set the right RGB colour to the LEDs
                    display_temp(temp);     //Display the read temperature
                }

            for(i=read_PWM_duty(BAR);i>0;i--)   //Ramp down bargraph
            {
              write_PWM_duty(BAR,i);
              Delay1KTCYx(50);
            }
            for(i=illum_level;i>0;i--){         //Ramp down illumination
              write_PWM_duty(BRIGHT,i);
              Delay1KTCYx(20);
            }

            stop_LED_pwm();

            SaveIntToEEPROM(SPUTTER_LOC,(LoadIntFromEEPROM(SPUTTER_LOC)-1));    //After 10,000 displays, clean up the tube and reset counter.
            if((LoadIntFromEEPROM(SPUTTER_LOC)-1)==0){
                sputter();
                SaveIntToEEPROM(SPUTTER_LOC,10000);
            }

            sleep_HV();          //Back to sleep
            minutes_slept=0;     //Reset sleep timers
            sleep_counter=0;
            TMR1H=0;
            TMR1L=0;
            T1_context=SLEEP;
            T1CONbits.TMR1ON=1; //start sleep alarm timer
            pb_up_flag=0;           //Catch presses of up button before going to sleep
       while(!((minutes_slept==disp_freq)|(pb_set_flag)|(pb_up_flag))){;}   //Sleep until it is time to wake up or a button is pressed
   }
}

/************************************************
* Function:		init
* Revision - Date:	2 - 14/3/2013
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	None
* Output:		None
* Side effects:		None
* Overview:		Initialisation of hardware
*
* Note:			None
*************************************************/
void init (void){

    //Oscillator configuration
    OSCCONbits.IRCF = 7;        //Select INTOSC 8MHz output
    //OSCTUNEbits.TUN = 3;        //Trim INTOSC frequency if necessary
    OSCTUNEbits.PLLEN = 1;      //Enable 4xPLL --> 4x8=32MHz clock. Accuracy over -10...+85degC is +/-5%

    //Voltage reference configuration
    CVRCONbits.CVRSS = 1;   //External Vref
    CVRCONbits.CVREN = 0;   //Bypass voltage reference

    //PWM configuration
    PTCON0bits.PTCKPS = 0;  //PWM time base prescaler set to 1:1, i.e. Fosc/4
    PTCON0bits.PTMOD = 0;   //PWM time base operates in free running mode. See chapter 13.3.1 in datasheet.
    PTPERH = 0;             //With Fosc=32MHz, Tpwm = 32usec hence fpwm = 31.25 kHz. Resolution becomes 10bit. See chapter 13.5 in datasheet.
    PTPERL = 255;
    PWMCON0bits.PMOD = 7;   //All PWM pins in independent mode
    write_PWM_duty(BRIGHT,0);    //Set all duty cycles to 0 i.e. inactive pins
    write_PWM_duty(HV,0);
    write_PWM_duty(BAR,0);
    PWMCON0bits.PWMEN = 7;  //All odd PWM I/O pins enabled for PWM output
    PTCON1bits.PTEN = 1;    //Start timebase

    //Comparator configuration
    CMCONbits.CMEN2=1;      //Enable Comparator 2

    //Timer 0 configuration
    T0CONbits.TMR0ON=0;     //Stop Timer 0. (It's ON after POR)
    T0CONbits.T016BIT=0;    //Timer 0 operates in 16-bit mode
    T0CONbits.T0CS=0;       //It is clocked by the instruction clock, 8MHz.
    T0CONbits.PSA=0;        //Clock source goes through prescaler
    T0CONbits.T0PS=0b010;   //Prescaler set to 1:8. Timer0 clock is therefore 1MHz.
    INTCON2bits.TMR0IP=0;   //Timer 0 interrupt has low priority
    INTCONbits.TMR0IE=1;    //Enable Timer 0 interrupts

    //Timer 1 configuration
    T1CONbits.RD16=1;       //Enable read/write in one 16bit operation
    T1CONbits.TMR1CS=0;     //It is clocked by the instruction clock, 8MHz.
    T1CONbits.T1CKPS=0b11;  //Prescaler to 1:8 (maximum). Timer 1 clock is therefore 1Mhz.
    IPR1bits.TMR1IP=1;      //Timer 1 has high priority
    PIE1bits.TMR1IE=1;      //Enable Timer 1 interrupts

    //Port direction configuration (Note, pins that are inputs by default are mentioned here too, except RA5)
    //PB_SET, PB_UP, RA2, RA3, T_SENSOR are inputs. (Note, T_SENSOR changes direction on runtime.
    //CMP2, RB3,  PGC are inputs.
    LATA = LATB = 0; //Set all outputs low.
    TRISAbits.RA6 = TRISAbits.RA7 = 0;  //RA6(CLKO), LEDB are outputs.
    TRISBbits.RB0 = TRISBbits.RB1 = TRISBbits.RB4 = TRISBbits.RB5 = TRISBbits.RB7 = 0;  //LED_R, LED_BRIGHTNESS, LED_G, HV_DRIVE, BARGRAPH_DRIVE are outputs.
    ADCON1 = 0x0F; //RA0, RA1, RA4 and RA6 set as inputs. Note, the description in the MCP datasheet is confusing.


    //Push button configuration
    INTCON2bits.INTEDG0=0;  //Push button SET interrupt on falling edge
    INTCON2bits.INTEDG1=0;  //Push button UP interrupt on falling edge
    INTCON3bits.INT1IP=1;   //Pushbutton UP is high priority (SET is high priority by hardware)

    //EEPROM configuration
    EECON1bits.EEPGD = 0;	//Set access to data EEPROM (see datasheet)
    EECON1bits.CFGS  = 0;

    //General Interrupt configuration
    RCONbits.IPEN=1;    //Enable interrupt priority levels
    INTCONbits.GIEH=1;  //Enable high priority interrupts
    INTCONbits.GIEL=1;  //Enable low priority interrupts

    if((PB_UP==0)&&(PB_SET==0)){    //If both buttons are pressed on startup
        wakeup_HV();
        calibration();              //Rough calibration
        sputter();                  //Tube regeneration
        calibration();              //Fine calibration
        sleep_HV();
    }

    //Temperature sensor configuration
    if(OW_reset()==0){                         //Reset and check for presence of sensor. If present,
        OW_write_byte(OW_SKIP_ROM);
        OW_write_byte(OW_WRITE_PAD);
        OW_write_byte(0x7F);                //TH byte = +127degC (effectively disabled)
        OW_write_byte(0x80);                //TL byte = -128degC (effectively disabled)
        OW_write_byte(0x00);                //Config byte = 9-bits temperature resolution
    }
    else{
        start_flash(RED);                  //If not present flash the red light
        while(1);                          //and do nothing as no sensor was detected!
    }

    INTCONbits.INT0IF=0;
    INTCON3bits.INT1IF=0;
    INTCON3bits.INT1E=1;    //Enable UP pushbutton interrupt
    INTCONbits.INT0IE=1;    //Enable SET pushbutton interrupt
}

/************************************************
* Function:		write_PWM_duty
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	Generator number, duty cycle
* Output:		None
* Side effects:		None
* Overview:		Writes ducty cycle value to
*                       designated PWM generator.
*
* Note:			None
*************************************************/
void write_PWM_duty(unsigned char gen, unsigned int duty){

    union Integer pwm;
    pwm.lt = duty;    // Copy duty cycle value into union

    switch(gen){
        case 0: PDC0H = pwm.bt[1];  // Write high byte to PWM generator 0
                PDC0L = pwm.bt[0];  // Write low byte to PWM generator 0
                break;
        case 1: PDC1H = pwm.bt[1];  // Write high byte to PWM generator 1
                PDC1L = pwm.bt[0];  // Write low byte to PWM generator 1
                break;
        case 2: PDC2H = pwm.bt[1];  // Write high byte to PWM generator 2
                PDC2L = pwm.bt[0];  // Write low byte to PWM generator 2
                break;
        default: break;
    }
}

/************************************************
* Function:		read_PWM_duty
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	Generator number
* Output:		Duty cycle
* Side effects:		None
* Overview:		Reads ducty cycle value from
*                       designated PWM generator.
*
* Note:			None
*************************************************/
unsigned int read_PWM_duty(unsigned char gen){

    union Integer pwm;

    switch(gen){
        case 0: pwm.bt[1]=PDC0H;  // Read high byte from PWM generator 0
                pwm.bt[0]=PDC0L;  // Read LOW byte from PWM generator 0
                break;
        case 1: pwm.bt[1]=PDC1H;
                pwm.bt[0]=PDC1L;
                break;
        case 2: pwm.bt[1]=PDC2H;
                pwm.bt[0]=PDC2L;
                break;
        default: break;
    }

    return pwm.lt;
}

/************************************************
* Function:		wakeup_HV
* Revision - Date:	2 - 14/3/2013
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	None
* Output:		None
* Side effects:		None
* Overview:		Takes the HV output voltage to
* approximately 140V after a slow ramp-up.
*
* Note:			Also see ISR. Will work without the wakeup function.
*************************************************/
void wakeup_HV(void){

    union Integer counts;
    unsigned char i;

    for (i=0;i<200;i++){
        write_PWM_duty(HV,i);   
        Delay10KTCYx(2);       //2.5ms delay between steps
    }
    counts.lt = 65115;          //Write Timer 0 counts value into union. 305 counts until overflow (65535). Approx 2.3kHz interrupt frequency.
    TMR0H=counts.bt[1];         //Write upper byte of counts value (pending writing of the lower byte below)
    TMR0L=counts.bt[0];         //Write lower byte of counts value. With this write the upper byte is also copied.
    T0CONbits.TMR0ON=1;         //Start the timer. Timer 0 interrupts will occur approx every 10ms.
    Delay10KTCYx(100);          //settling delay
}

/************************************************
* Function:		sleep_HV
* Revision - Date:	2 - 14/3/2013
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	None
* Output:		None
* Side effects:		None
* Overview:		Switches the boost converter off.
*
* Note:			Output capacitor discharge time.
*************************************************/
void sleep_HV(void){

    T0CONbits.TMR0ON=0;        //Stop Timer 0
    write_PWM_duty(HV,0);      //Set HV converter to OFF
}

/************************************************
* Function:		calibration
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	None
* Output:		None
* Side effects:		None
* Overview:		Output a specific duty cycle to the
* bargraph driver. The user must then use the on-board trimmer (P1)
* to set the height of the glowing bar to 100 deg F. This forces a known
* mm/mA constant, compensating for differences between bargraph tubes.
* Waits for SET pushbutton to exit.
*
* Note:			Also see the temp_to_pwm function
*************************************************/
void calibration(void){

    write_PWM_duty(BAR,743);        //This PWM register value (72.6% duty cycle) generates 3.6V for the current sink of the indicator tube. The trimmer must be set for the tube to indicate 100degF.
    while((PB_UP==0)&&(PB_SET==0)); //Wait for buttons to be depressed
    Delay10KTCYx(100);              //approx. 125ms additional debouncing delay
    while(PB_SET==1);               //Wait until SET pushbutton is pressed to exit
    write_PWM_duty(BAR,0);
}

/************************************************
* Function:		start_flash
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	Flash colour, RED or GREEN
* Output:		None
* Side effects:		Claims hardware Timer 1
* Overview:		Flashes either the green or red LEDs
* to help menu navigation.
* Note:			Also see ISR.
*************************************************/
void start_flash(unsigned char colour){

    if(colour==RED)LED_R=1;         //Actvate desired led colour
    else LED_G=1;
    write_PWM_duty(BRIGHT,700);     //Set a resonable illumination level
    T1_context=FLASH;
    TMR1H=0;                        //Set Timer 1 for the full 2^16 or 65536 steps
    TMR1L=0;
    T1CONbits.TMR1ON=1;             //Start the timer. Interrupts will occur approx every 6.6ms
}

/************************************************
* Function:		stop_flash
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	None
* Output:		None
* Side effects:		None
* Overview:		Stops flashing of green or red LEDs
* and sets their brightness to 0.
* Note:			None
*************************************************/
void stop_flash(void){

    T1CONbits.TMR1ON=0;             //Stop the timer
    LED_R=0;                        //Switch off both LEDs (only 1 of the two will be on)
    LED_G=0;
    write_PWM_duty(BRIGHT,0);       //set the brightness to 0
}

/************************************************
* Function:		temp_to_pwm
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	temperature in degrees C (double the real temperature)
* Output:		None
* Side effects:		None
* Overview:		Converts a temperature reading to a PWM value
* for the bargraph driver to display. A value twice the real temperature is
* used to achieve 0.5 degC resolution without floating point arithmetic.
* If the temperature is outside the range of the scale, fixed minimum and maximum values
* are returned. Two functions are used to compensate for the increasing non-linearity (mm/mA)
* of the bargraph tube towards the top of its length. The negative value subtracted arises from the
* fact that the scale starts from 10 degC, not 0.
* Note:			None
*************************************************/
unsigned int temp_to_pwm (unsigned char temp){

    unsigned int pwm;

    if(temp<20)pwm=0;   //If the temperature is less than 10 degC, return a value to display 10 degC.
    else if(temp>82) pwm=840;   //If it is more than 41degC, return a value to display 41 degC.
    else if(temp<70){           //Else if it is less than 35 solve this function:
        pwm=(unsigned int)(temp*(unsigned int)13)-(unsigned int)248;    //The function solved here is PWM(equivalent to mm of height) = temperature * 26 - 248.
    } else{ //For more than 35, solve this function:
        pwm=(unsigned int)(temp*(unsigned int)16)-(unsigned int)461;    //The function solved here is PWM(equivalent to mm of height) = temperature * 32 - 461.
    }
    return pwm;
}

/************************************************
* Function:		display_temp
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	temperature in degrees C (double the real temperature)
* Output:		None
* Side effects:		None
* Overview:		Displays the temperature without any effects.
* Note:			Also see the display_temp_effects function
*************************************************/
void display_temp(unsigned char temp){
    write_PWM_duty(BAR,temp_to_pwm(temp));      //Simply write the calculated PWM register value to the bargraph PWM register
}

/************************************************
* Function:		enter_menu
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	None
* Output:		None
* Side effects:		Claims hardware Timer 1.
* Overview:		Displays top menu with options.
* Going though all selections without entering one will exit the menu.
* Note:			Also see the enter_menu_option function
*************************************************/
void enter_menu(void){

    unsigned char menu_pointer=1;   //Set menu pointer to the first item
    unsigned char option_temp=20;   //Temperature to display for first item is 10degC

    start_flash(GREEN);             //Start green flash to indicate top level
    display_temp(option_temp);

    while(menu_pointer<8){          //While the menu item is less than 8
        if(pb_up_flag){             //If up is pressed
            pb_up_flag=0;
            ++menu_pointer;         //go to next menu item
            option_temp += 10;      //increase displayed temeprature by 5deg C
            display_temp(option_temp);}
        else if(pb_set_flag){       //If set flag is pressed
            pb_set_flag=0;
            stop_flash();           //stop green flash
            enter_menu_option(menu_pointer);    //enter configuration menu for current menu item
            start_flash(GREEN);                 //Once done, flash green again
            display_temp(option_temp);}         //and display current menu item temperature
    }
    write_PWM_duty(BAR,0);
    stop_flash();
}

/************************************************
* Function:		enter_menu_option
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	Selected option (1 to 7)
* Output:		None
* Side effects:		Claims hardware Timer 1.
* Overview:		Displays configuration settings for
* selected option. Pressing SET saves the setting. All settings are
* saved in EEPROM before exit.
*
* Note:			Also see the enter_menu function
*************************************************/
void enter_menu_option(unsigned char option){
    start_flash(RED);
    write_PWM_duty(BAR,0);
    Delay10KTCYx(30);   //Wait for tube to de-ionize

    switch(option){     //Depending on which menu item, select implementation
        //Display duration
        case(1):
            do{
                switch(disp_dur){
                    case(5):
                                display_temp(20);
                                if(pb_up_flag){
                                    pb_up_flag=0;
                                    disp_dur=10;
                                }
                                break;
                    case(10):   display_temp(30);
                                if(pb_up_flag){
                                    pb_up_flag=0;
                                    disp_dur=15;
                                }
                                break;
                    case(15):   display_temp(40);
                                if(pb_up_flag){
                                    pb_up_flag=0;
                                    disp_dur=20;
                                }
                                break;
                    case(20):   display_temp(50);
                                if(pb_up_flag){
                                    pb_up_flag=0;
                                    disp_dur=30;
                                }
                                break;
                    case(30):   display_temp(60);
                                if(pb_up_flag){
                                    pb_up_flag=0;
                                    disp_dur=45;
                                }
                                break;
                    case(45):   display_temp(70);
                                if(pb_up_flag){
                                    pb_up_flag=0;
                                    disp_dur=60;
                                }
                                break;
                    case(60):   display_temp(80);
                                if(pb_up_flag){
                                    pb_up_flag=0;
                                    disp_dur=5;
                                }
                                break;
                    default:break;
                }
            }while(pb_set_flag==0);
            pb_set_flag=0;
            break;
        //Frequency of display
        case(2):
            do{
                switch(disp_freq){
                    case(1):
                                display_temp(20);
                                if(pb_up_flag){
                                    pb_up_flag=0;
                                    disp_freq=2;
                                }
                                break;
                    case(2):   display_temp(30);
                                if(pb_up_flag){
                                    pb_up_flag=0;
                                    disp_freq=5;
                                }
                                break;
                    case(5):   display_temp(40);
                                if(pb_up_flag){
                                    pb_up_flag=0;
                                    disp_freq=10;
                                }
                                break;
                    case(10):   display_temp(50);
                                if(pb_up_flag){
                                    pb_up_flag=0;
                                    disp_freq=20;
                                }
                                break;
                    case(20):   display_temp(60);
                                if(pb_up_flag){
                                    pb_up_flag=0;
                                    disp_freq=30;
                                }
                                break;
                    case(30):   display_temp(70);
                                if(pb_up_flag){
                                    pb_up_flag=0;
                                    disp_freq=60;
                                }
                                break;
                    case(60):   display_temp(80);
                                if(pb_up_flag){
                                    pb_up_flag=0;
                                    disp_freq=1;
                                }
                                break;
                    default:break;
                }
            }while(pb_set_flag==0);
            pb_set_flag=0;
            break;
        //Cold setpoint
        case(3):
                display_temp(2*cold_set);   //Double the cold setpoint temperature to display it correctly
                while(pb_set_flag==0){
                    while(PB_UP==0){
                        cold_set++;         //Increment setpoint while UP is being pressed
                        if(cold_set>hot_set){   //If cold setpoint has reached the hot setpoint
                            write_PWM_duty(BAR,0);  //set the bargraph to 0
                            Delay10KTCYx(30);      //Wait for tube to de-ionize
                            cold_set=10;           //Start over
                        }
                        display_temp(2*cold_set);
                        Delay10KTCYx(200);          //A small delay to slow down incrementing
                    }
                }
                pb_up_flag=0;
                pb_set_flag=0;
                break;
        //Hot setpoint
        case(4):
                display_temp(2*hot_set);
                while(pb_set_flag==0){
                    while(PB_UP==0){
                        hot_set++;
                        if(hot_set>40){
                            write_PWM_duty(BAR,0);
                            Delay10KTCYx(30);       //Wait for tube to de-ionize
                            hot_set=cold_set+1;
                        }
                        display_temp(2*hot_set);
                        Delay10KTCYx(200);          //A small delay to slow down incrementing
                    }
                }
                pb_up_flag=0;
                pb_set_flag=0;
                break;
        //Cold setpoint colour
        case(5):
                do{
                    switch(cold_hue){
                        case(RED_HUE_H):
                        case(RED_HUE_L):
                                        display_temp(20);
                                        if(pb_up_flag){
                                            pb_up_flag=0;
                                            cold_hue=MAGENTA_HUE;
                                        }
                                        break;
                        case(MAGENTA_HUE):display_temp(30);
                                    if(pb_up_flag){
                                        pb_up_flag=0;
                                        cold_hue=BLUE_HUE;
                                    }
                                    break;
                        case(BLUE_HUE):display_temp(40);
                                    if(pb_up_flag){
                                        pb_up_flag=0;
                                        cold_hue=CYAN_HUE;
                                    }
                                    break;
                        case(CYAN_HUE):   display_temp(50);
                                    if(pb_up_flag){
                                        pb_up_flag=0;
                                        cold_hue=GREEN_HUE;
                                    }
                                    break;
                        case(GREEN_HUE):display_temp(60);
                                    if(pb_up_flag){
                                        pb_up_flag=0;
                                        cold_hue=YELLOW_HUE;
                                    }
                                    break;
                        case(YELLOW_HUE):display_temp(70);
                                    if(pb_up_flag){
                                        pb_up_flag=0;
                                        cold_hue=WHITE;
                                    }
                                    break;
                        case(WHITE):  display_temp(80);
                                    if(pb_up_flag){
                                        pb_up_flag=0;
                                        cold_hue=RED_HUE_L;
                                    }
                                    break;
                        default:break;
                    }
                }while(pb_set_flag==0);
                pb_set_flag=0;
                break;
        //Hot hue setpoint
        case(6):
                do{
                    switch(hot_hue){
                        case(RED_HUE_H):
                        case(RED_HUE_L):
                                        display_temp(20);
                                        if(pb_up_flag){
                                            pb_up_flag=0;
                                            hot_hue=MAGENTA_HUE;
                                        }
                                        break;
                        case(MAGENTA_HUE):display_temp(30);
                                    if(pb_up_flag){
                                        pb_up_flag=0;
                                        hot_hue=BLUE_HUE;
                                    }
                                    break;
                        case(BLUE_HUE):display_temp(40);
                                    if(pb_up_flag){
                                        pb_up_flag=0;
                                        hot_hue=CYAN_HUE;
                                    }
                                    break;
                        case(CYAN_HUE):   display_temp(50);
                                    if(pb_up_flag){
                                        pb_up_flag=0;
                                        hot_hue=GREEN_HUE;
                                    }
                                    break;
                        case(GREEN_HUE):display_temp(60);
                                    if(pb_up_flag){
                                        pb_up_flag=0;
                                        hot_hue=YELLOW_HUE;
                                    }
                                    break;
                        case(YELLOW_HUE):display_temp(70);
                                    if(pb_up_flag){
                                        pb_up_flag=0;
                                        hot_hue=WHITE;
                                    }
                                    break;
                        case(WHITE):  display_temp(80);
                                    if(pb_up_flag){
                                        pb_up_flag=0;
                                        hot_hue=RED_HUE_L;
                                    }
                                    break;
                        default:break;
                    }
                }while(pb_set_flag==0);
                pb_set_flag=0;
                break;
        //Illumination level
        case(7):
                do{
                    switch(illum_level){
                        case(OFF):
                                    display_temp(20);
                                    if(pb_up_flag){
                                        pb_up_flag=0;
                                        illum_level=LVL_1;
                                    }
                                    break;
                        case(LVL_1):display_temp(30);
                                    if(pb_up_flag){
                                        pb_up_flag=0;
                                        illum_level=LVL_2;
                                    }
                                    break;
                        case(LVL_2):display_temp(40);
                                    if(pb_up_flag){
                                        pb_up_flag=0;
                                        illum_level=LVL_3;
                                    }
                                    break;
                        case(LVL_3):   display_temp(50);
                                    if(pb_up_flag){
                                        pb_up_flag=0;
                                        illum_level=LVL_4;
                                    }
                                    break;
                        case(LVL_4):display_temp(60);
                                    if(pb_up_flag){
                                        pb_up_flag=0;
                                        illum_level=LVL_5;
                                    }
                                    break;
                        case(LVL_5):display_temp(70);
                                    if(pb_up_flag){
                                        pb_up_flag=0;
                                        illum_level=MAX;
                                    }
                                    break;
                        case(MAX):  display_temp(80);
                                    if(pb_up_flag){
                                        pb_up_flag=0;
                                        illum_level=OFF;
                                    }
                                    break;
                        default:break;
                    }
                }while(pb_set_flag==0);
                pb_set_flag=0;
                break;
        default:break;
    }
    write_PWM_duty(BAR,0);
    Delay10KTCYx(30);
    stop_flash();
    save_user_settings();
}

// <editor-fold defaultstate="collapsed" desc="1-wire functions">
/************************************************
* Function:		OW_reset
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	None
* Output:		Presence of device
* Side effects:		None
* Overview:		Resets the 1-wire bus and checks for presence
 * of at least one device. Returns 0 if a device is present, 1 if not.
* Note:			None
*************************************************/
unsigned char OW_reset(void) {

    unsigned char presence;
    OW_PORT=0;                 //Take the bus low
    OW_RW=WRITE;
    Delay100TCYx(44);          //550us@8MHz, accounting for clock accuracy and safety margin
    OW_RW=READ;                //Release the bus
    Delay10TCYx(54);           //67.5us@8MHz, about the centre of the worst case presence pulse
    presence = OW_PORT;        //Read the bus
    Delay100TCYx(39);          //487us@8MHz, clear the potential remaining time of the presence pulse
    return presence;           //0=1-wire device present, 1=1-wire device not found
}

/************************************************
* Function:		OW_write_bit
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	Bit to write
* Output:		None
* Side effects:		None
* Overview:		Writes 1 bit of data to the 1-wire bus
* Note:			None
*************************************************/
void OW_write_bit(unsigned char data) {
    if (data) {                //Execute a '1 time slot', writing 1 to the bus
        OW_PORT=0;             //Take the bus low to start time slot
        OW_RW=WRITE;
        Delay10TCYx(4);        //5us@8MHz, more than 1us required, less than 15us potential bus read time from slave device
        OW_RW=READ;            //Release the bus
        Delay100TCYx(6);       //70us@8MHz, allow sufficient time for the slave to read the bus
    }
    else {                     //Execute a '0 time slot', writing 0 to the bus
        OW_PORT=0;             //Take the bus low to start time slot
        OW_RW=WRITE;
        Delay100TCYx(6);       //70us@8MHz, keep the bus low for the slave to read it
        OW_RW=READ;            //Release the bus
    }
    Delay10TCYx(4);            //5us@8MHz, recovery delay between time slots
}

/************************************************
* Function:		OW_write_byte
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	Byte to write
* Output:		None
* Side effects:		None
* Overview:		Writes 1 byte of data to the 1-wire bus
* Note:			None
*************************************************/
void OW_write_byte(unsigned char data) {
    unsigned char i;
    for (i=0;i<8;i++) {                 //repeat for all 8 bits
        OW_write_bit(data & 0x01);      //mask for only the LSB of the byte
        data >>= 1;                     //shift the bits in the byte one position to the right thus preparing to send the next bit
    }
}

/************************************************
* Function:		OW_read_bit
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	None
* Output:		Bit read
* Side effects:		None
* Overview:		Reads 1 bit of data from the 1-wire bus
* Note:			None
*************************************************/
unsigned char OW_read_bit(void) {
    unsigned char data;

    OW_PORT=0;             //Take the bus low to start time slot
    OW_RW=WRITE;
    Delay10TCYx(2);        //2.5us@8MHz, more than 1us required, maximise timing margin
    OW_RW=READ;            //Release the bus
    Delay10TCYx(8);        //10us@8MHz, 10+2.5us = 12.5us, towards the end of the 15us read period with margin
    data = OW_PORT;        //read the bus
    Delay100TCYx(6);       //70us@8MHz, allow sufficient time for the slave to release the bus
    return data;
}

/************************************************
* Function:		OW_read_byte
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	None
* Output:		Byte read
* Side effects:		None
* Overview:		Reads 1 byte of data from the 1-wire bus
* Note:			None
*************************************************/
unsigned char OW_read_byte(void) {
    unsigned char data=0;
    unsigned char i;

    for (i=0;i<8;i++) {             //repeat for all 8 bits
        data >>= 1;                 //shift the byte to the right to prepare for the next bit
        if (OW_read_bit()) data = data | 0x80;  //if the bit is high, set the MSB of the byte (otherwise it will remain 0)
    }
    return data;
}
// </editor-fold>

/************************************************
* Function:		convert_temp
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	None
* Output:		None
* Side effects:		None
* Overview:		Starts a temperature conversion
* Note:			Assumes 1 device connected using parasite power
*************************************************/
void convert_temp(void) {
    OW_reset();
    OW_write_byte(OW_SKIP_ROM);     //Skip ROM
    OW_write_byte(OW_CONVERT_T);    //Start temperature conversion (using current scratchpad settings)
    OW_PORT=1;                      //Enable strong pull-up on DQ
    OW_RW=WRITE;
    Delay10KTCYx(80);               //100ms@8MHz conversion delay
    OW_RW=READ;                     //Release bus
}

/************************************************
* Function:		read_temp
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	At least one conversion must
* take place for temperature to be valid.
* Input:        	None
* Output:		Temperature read
* Side effects:		None
* Overview:		Reads the value of temperature from the sensor
* Note:			For negative temperatures the value returned is 0.
*************************************************/
unsigned int read_temp(void){

    union Integer temp;

    OW_reset();
    OW_write_byte(OW_SKIP_ROM);  // Skip ROM selection
    OW_write_byte(OW_READ_PAD);  // Read scratch pad RAM

    temp.bt[0]=OW_read_byte();   //Read the first byte = LSB of temperature
    temp.bt[1]=OW_read_byte();   //Read the second byte = MSB of temperature

    OW_reset(); //Issue a reset to terminate the reading process as nothing else is needed

    if(temp.lt > 63488)return 0;     //If the integer read has the 5 MSBs (sign bits) set, the temperature is negative, set it to 0. 5 MSBs set for an unsigned int is 63488.
    else return (temp.lt >>=3);    // The 3 LSBs are undefined for 9-bit resolution, so shift right 3 times. The significance of the LSB is 0.5deg so the value returned is 2x the real temperature
}

/************************************************
* Function:		delay_minx
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	Minutes x to wait
* Output:		None
* Side effects:		None
* Overview:		A software delay for 0...255 minutes
* Note:			Not using a hardware timer as these are used for PWM of LEDs
*************************************************/
void delay_minx(unsigned char minutes){
    unsigned char i,j;
    for(j=minutes;j>0;j--){
        for(i=0;i<188;i++)
        {Delay10KTCYx(255);}    //1 min @8MIPS
    }
}

/************************************************
* Function:		delay_secx
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	Seconds x to wait
* Output:		None
* Side effects:		None
* Overview:		A software delay for 0...255 seconds. Pressing the SET button will end the delay.
* Note:			None
*************************************************/
void delay_secx(unsigned char seconds){
    unsigned char i,j;
    for(j=seconds;j>0;j--){
        for(i=0;i<200;i++)      //1 sec @8MIPS
        {
            Delay10KTCYx(4);
            if(pb_set_flag)goto end;    //If during the delay the SET button has been pressed, stop the delay
        }
    }
end:;
}

/************************************************
* Function:		sputter
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	HV supply must be awake
* Input:        	None
* Output:		None
* Side effects:		None
* Overview:		Overdrives the tube for 2 minutes for cleaning. Pushbuttons are unresponsive during that period.
* Note:			Do not overuse as tube will become darker.
*************************************************/
void sputter(void){

    write_PWM_duty(BAR,1023);   //Set the bargraph drive to maximum (overdrive the tube)
    delay_minx(2);              //Allow two minutes cleaning time
    write_PWM_duty(BAR,0);      //Switch it off
}

// <editor-fold defaultstate="collapsed" desc="EEPROM functions">
/************************************************
* Function:		EEPROMWrite
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	Address to write to, Data byte to write
* Output:		None
* Side effects:		Disables all interrupts during writing
* Overview:		Writes a byte of data to the specified address
* Note:			None
*************************************************/
void EEPROMWrite(unsigned char address, unsigned char data){

 INTCONbits.GIEH=0;  //Disable high priority interrupts
 INTCONbits.GIEL=0;  //Disable low priority interrupts

 EEADR	= address;
 EEDATA = data;						//Set the data to be written
 EECON1bits.WREN = 1;				//Enables a write to the EEPROM

 EECON2 = 0x55;						//This sequence is specified in the MCU datasheet and must be followed. EECON2 is not a physical register
 EECON2 = 0x0AA;
 EECON1bits.WR = 1;					//Initialise a write cycle

 while(EECON1bits.WR==1){;}			//Wait for the write cycle to complete

 EECON1bits.WREN = 0;				//Disable writes to the EEPROM

 INTCONbits.GIEH=1;  //Enable high priority interrupts
 INTCONbits.GIEL=1;  //Enable low priority interrupts
}

/************************************************
* Function:		EEPROMRead
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	Address to read from
* Output:		Data byte read
* Side effects:		None
* Overview:		Reads a byte of data from the specified address
* Note:			None
*************************************************/
unsigned char EEPROMRead(unsigned char address){
 EEADR	= address;
 EECON1bits.RD = 1;					//Initialise a read cycle
 return EEDATA;						//return the result which is available in the next instruction cycle and is stored in the EEDATA register
}

/************************************************
* Function:		SaveByteToEEPROM
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	Address to write to, Data byte to write
* Output:		None
* Side effects:		Disables all interrupts during writing
* Overview:		Writes a byte of data to the specified address
* Note:			None
*************************************************/
void SaveByteToEEPROM(unsigned char address, unsigned char data){
 EEPROMWrite(address,data);		//Call the simple read from EEPROM
}

/************************************************
* Function:		SaveIntToEEPROM
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	Address to write to, Data integer to write
* Output:		None
* Side effects:		Disables all interrupts during writing
* Overview:		Writes an integer of data to the specified address
* Note:			Assumes little-endian system
*************************************************/
void SaveIntToEEPROM(unsigned char start_address, unsigned int data){
 char *pdata =  &data;      	//Declare a pointer of type char(1Byte) pointing to the first byte of the input data

 EEPROMWrite(start_address, *pdata);          //Write the first byte of data
 pdata++;                                  	  //increment pointer to next Byte
 EEPROMWrite((start_address+1), *pdata);      //Write the second byte of data
}

/************************************************
* Function:		LoadByteFromEEPROM
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	Address to read from
* Output:		Data byte read
* Side effects:		None
* Overview:		Reads a byte of data from the specified address
* Note:			None
*************************************************/
unsigned char LoadByteFromEEPROM(unsigned char address){
 return EEPROMRead(address);		//Call the simple read from EEPROM and return the result
}

/************************************************
* Function:		LoadIntFromEEPROM
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	Address to read from
* Output:		Data integer read
* Side effects:		None
* Overview:		Reads an integer of data from the specified address
* Note:			Assumes little-endian system.
*************************************************/
unsigned int LoadIntFromEEPROM(unsigned char start_address){
 unsigned int result=0;							//Declare a result variable to work with
 char *presult =  &result;						//Declare a pointer of character type (8bits) and have it pointing to the beginning of the result variable above

 *presult = EEPROMRead(start_address);			//Read the first byte of data
  presult++;									//increment pointer to next Byte
 *presult = EEPROMRead((start_address+1));		//Read the second byte of data

 return result;									//Return the result
}
// </editor-fold>

/************************************************
* Function:		load_user_settings
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	None
* Output:		None
* Side effects:		None
* Overview:		Loads user settings from EEPROM
* Note:			None
*************************************************/
void load_user_settings(void){
    disp_freq=LoadByteFromEEPROM(DISP_FREQ_LOC);
    disp_dur=LoadByteFromEEPROM(DISP_DUR_LOC);
    hot_set=LoadByteFromEEPROM(HOT_SET_LOC);
    cold_set=LoadByteFromEEPROM(COLD_SET_LOC);
    hot_hue=LoadIntFromEEPROM(HOT_HUE_LOC);
    cold_hue=LoadIntFromEEPROM(COLD_HUE_LOC);
    illum_level=LoadIntFromEEPROM(ILLUM_LVL_LOC);
}

/************************************************
* Function:		save_user_settings
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	None
* Output:		None
* Side effects:		All interrupts are disabled during write
* Overview:		Saves all settings to EEPROM. Also sets
* EEPROM location 0 to 0, which indicates that user settings now exist in memory.
* Note:			None
*************************************************/
void save_user_settings(void){
    SaveByteToEEPROM(DISP_FREQ_LOC, disp_freq);
    SaveByteToEEPROM(DISP_DUR_LOC, disp_dur);
    SaveByteToEEPROM(HOT_SET_LOC, hot_set);
    SaveByteToEEPROM(COLD_SET_LOC, cold_set);
    SaveIntToEEPROM(HOT_HUE_LOC, hot_hue);
    SaveIntToEEPROM(COLD_HUE_LOC, cold_hue);
    SaveIntToEEPROM(ILLUM_LVL_LOC, illum_level);
}

/************************************************
* Function:		start_LED_pwm
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	None
* Output:		None
* Side effects:		Claims Timer 1
* Overview:		Starts PWM of RGB LEDs based on red green and blue values (see variables)
* Note:			Also see ISR.
*************************************************/
void start_LED_pwm(void){

    union Integer counts;
    counts.lt = 64910;          //Write Timer 1 counts value into union. 1250 counts until overflow (65535). Approx 800Hz interrupt frequency.
    TMR1H=counts.bt[1];         //Write upper byte of counts value (pending writing of the lower byte below)
    TMR1L=counts.bt[0];         //Write lower byte of counts value. With this write the upper byte is also copied.
    T1_context=PWM;             //Set the right context for Timer 1
    led_PWM_counter=0;
    if(red_duty>0)LED_R=1;
    if(green_duty>0)LED_G=1;
    if(blue_duty>0)LED_B=1;
    T1CONbits.TMR1ON=1;         //Start the timer. Timer 1 interrupts will occur approx every 1.25ms.
}

/************************************************
* Function:		stop_LED_pwm
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	None
* Output:		None
* Side effects:		None
* Overview:		Stops RGB LED PWM and sets brightness to 0.
* Note:			Also see ISR.
*************************************************/
void stop_LED_pwm(void){
    T1CONbits.TMR1ON=0; //Stop the timer
    LED_R=0;            //Switch off LEDs
    LED_G=0;
    LED_B=0;
}

/************************************************
* Function:		hue_to_RGB
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	Hue- 0...360
* Output:		Red, Green, Blue intensity
* Side effects:		None
* Overview:		Calculates the RGB values for a given Hue.
* This function is a simplified version of a HSV to RGB function; saturation and value being set to 1(max).
* RGB output values are scaled from 0 to 15.
* Note:			Also read http://en.wikipedia.org/wiki/HSL_and_HSV#From_HSV
*************************************************/
void hue_to_RGB(unsigned int H, unsigned char *R, unsigned char *G, unsigned char *B){

    unsigned int segment;
    unsigned long x;

    if ((H>=360)||(H<0)) H=0;   //If the hue is out of range, set it to 0 (red).

    segment =(H*10)/6;          //Determine one which segment of the colour rgb colour circle we are
    x=10000L*(100-abs((segment%200)-100)); //Calculate value of less dominant colour

    switch (H/60){      //Depending on segment, assign dominant colour intensity to max, less dominant, and least dominant to 0.
        case 0:
                *R=15;
                *G=(15*x)/1000000;
                *B=0;
                break;
        case 1:
                *R=(15*x)/1000000;
                *G=15;
                *B=0;
                break;
        case 2:
                *R=0;
                *G=15;
                *B=(15*x)/1000000;;
                break;
        case 3:
                *R=0;
                *G=(15*x)/1000000;
                *B=15;
                break;
        case 4:
                *R=(15*x)/1000000;
                *G=0;
                *B=15;
                break;
        case 5:
                *R=15;
                *G=0;
                *B=(15*x)/1000000;
                break;
    }

}

/************************************************
* Function:		abs
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	number
* Output:		Absolute value of number
* Side effects:		None
* Overview:		Calculates the absolute value of a number
* Note:			None
*************************************************/
unsigned int abs(signed int value){
if (value < 0)
return -value;
else
return value;
}

/************************************************
* Function:		find_hue
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	Hue of cold and hot setpoints, temperature of hot and cold setpoints, current temperature
* Output:		Hue at current temperature
* Side effects:		None
* Overview:		Calculates the hue for scale illumination at current temperature, using hold and cold preset colours at cold and cold preset temperature setpoints.
* Note:			None
*************************************************/
unsigned int find_hue (unsigned int hot_hue, unsigned int cold_hue, unsigned char hot_set, unsigned char cold_set, unsigned char temperature){

    unsigned char direction;
    unsigned int hue_diff;
    unsigned int a;
    signed long b;
    unsigned int hue_os;

    temperature=temperature/2;                          //Divide the temperature by two so that it can be compared with the setpoints
    if (hot_hue==cold_hue)return hot_hue;               //If the same colour has been set, return that colour
    else if (temperature>=hot_set) return hot_hue;      //If the temperature is above the hot setpoint, return the colour of the hot setpoint
    else if (temperature<=cold_set) return cold_hue;    //Similarly for the cold setpoint
    else{
        switch(hot_hue){                                //Otherwise depending on the value of the hue for the hot setpoint, determine which directing we are moving on the RGB colour wheel (clock wise or counterclockwise)
            case RED_HUE_H:
            case RED_HUE_L:
                switch(cold_hue){
                    case MAGENTA_HUE:                   //If the hot setpoint colour is red, then if the cold setpoint colour is magenta, blue or cyan
                    case BLUE_HUE:
                    case CYAN_HUE:
                        direction=1;                    //We are moving up the hue values (counterclockwise)
                        hot_hue=RED_HUE_H;              //and red has a huw value of 360 (instead of 0, needed for calculations later)
                        break;
                    case GREEN_HUE:
                    case YELLOW_HUE:
                        direction=0;
                        hot_hue=RED_HUE_L;
                        break;
                }
                break;
            case MAGENTA_HUE:
                 switch(cold_hue){
                    case RED_HUE_H:
                    case RED_HUE_L:
                        direction=0;
                        hot_hue=RED_HUE_H;
                        break;
                    case BLUE_HUE:
                    case CYAN_HUE:
                    case GREEN_HUE:
                    case YELLOW_HUE:
                        direction=1;
                        break;
                }
                break;
            case BLUE_HUE:
                 switch(cold_hue){
                    case RED_HUE_H:
                    case RED_HUE_L:
                        direction=0;
                        hot_hue=RED_HUE_H;
                        break;
                    case MAGENTA_HUE:
                        direction=0;
                        break;
                    case CYAN_HUE:
                    case GREEN_HUE:
                    case YELLOW_HUE:
                        direction=1;
                        break;
                }
                break;
           case CYAN_HUE:
                 switch(cold_hue){
                    case RED_HUE_H:
                    case RED_HUE_L:
                        direction=0;
                        hot_hue=RED_HUE_H;
                        break;
                    case MAGENTA_HUE:
                    case BLUE_HUE:
                        direction=0;
                        break;
                    case GREEN_HUE:
                    case YELLOW_HUE:
                        direction=1;
                        break;
                }
                break;
           case GREEN_HUE:
                 switch(cold_hue){
                    case RED_HUE_H:
                    case RED_HUE_L:
                        direction=1;
                        hot_hue=RED_HUE_L;
                        break;
                    case MAGENTA_HUE:
                    case BLUE_HUE:
                    case CYAN_HUE:
                        direction=0;
                        break;
                    case YELLOW_HUE:
                        direction=1;
                        break;
                }
                break;
           case YELLOW_HUE:
                 switch(cold_hue){
                    case RED_HUE_H:
                    case RED_HUE_L:
                        direction=1;
                        hot_hue=RED_HUE_L;
                        break;
                    case MAGENTA_HUE:
                    case BLUE_HUE:
                    case CYAN_HUE:
                    case GREEN_HUE:
                        direction=0;
                        break;
                }
                break;
        }

        hue_diff = abs(hot_hue-cold_hue); //find the hue difference between the two colours
        a=1000/(hot_set-cold_set);      //Find scale (a) and offset (b) to solve the equation below.
        b=(signed long)(a*cold_set);
        hue_os=(a*temperature)-b;       //Solve equation to find how much the setpoint hue must be offset for the current temperature
        hue_os = (unsigned int)(((unsigned long)hue_os*hue_diff)/1000L);    //Now scale the hue difference by the offset factor from above. The result is a hue offset from either the cold or hot setpoints

        if(direction==1)return (cold_hue+hue_os+5); //If going up the hue values, add the hue offset to the cold hue. Also account for some error due to led voltage control.
        else return (cold_hue-hue_os-5); //If going down, subract the offset.
    }
}

/************************************************
* Function:		pre_scale_colour
* Revision - Date:	1 - 2/6/2012
* Author:		SALTechips - Alex Tsekenis
* Compiler:		C18 v3.37.01
* Pre-condition: 	None
* Input:        	None
* Output:		None
* Side effects:		None
* Overview:		Sets the colour of the scale illumination depending on user settings and current temperature
* Note:			None
*************************************************/
void prep_scale_colour(void){
    if((cold_hue==WHITE)|(hot_hue==WHITE)){    //If the 'hue' of the cold setpoint has been set to white (which means the hot is also white)
             red_duty=15;   //Set the intensity of all LEDs to maximum
             green_duty=15;
             blue_duty=15;
     }
    else hue_to_RGB(find_hue(hot_hue,cold_hue,hot_set,cold_set,temp), &red_duty, &green_duty, &blue_duty); //Otherwise calculate the hue at the current temperature and then calculate the right RGB values.
}